// capture.ts
// main entry point for Socket Mobile CaptureJS SDK
//
// Copyright © 2019 Socket Mobile, Inc.
// ignore_for_file: unnecessary_this, unnecessary_null_comparison

import 'dart:async';
import 'dart:io';
import 'dart:convert';

import 'package:flutter/material.dart';

import './index.dart';
import './gen/errors.dart';

const errMsgNoTransport = 'no transport, is this initialized?';

String defaultHost = "http://127.0.0.1:18481";

class Capture {
  Transport? transport;
  String host = defaultHost;
  int? clientOrDeviceHandle;
  int? transportHandle;
  Function? onEventNotification;
  Capture? rootCapture;
  Logger? logger;
  bool? iOS;

  Capture([this.logger, this.clientOrDeviceHandle, this.transportHandle, this.onEventNotification, this.transport]);

  Map<String, dynamic> toJson() => {
        'transport': transport,
        'host': host,
        'clientOrDeviceHandle': clientOrDeviceHandle,
        'transportHandle': transportHandle, 
        'rootCapture': rootCapture
  };

  transportHelper(AppInfo appInfo, Function eventNotification, [CaptureOptions? options]) async {
    if (options != null){
      this.transport = options.transport ?? Transport().getTransport(this.logger);
      this.host = options.host ?? defaultHost;
    } else {
      this.transport = Transport().getTransport(this.logger);
    }

    try {
      this.clientOrDeviceHandle = await this.transport?.open(host, appInfo, (dynamic event, int handle) { 
        return eventNotification(event, handle);
      });
      return this.clientOrDeviceHandle;
    } catch (res){
      return JRpcAndIosError(0, -32602, res.toString()).toJson();
    }

  }

  Future open(AppInfo appInfo, Function eventNotification, [CaptureOptions? options]) {
    return transportHelper(appInfo, eventNotification, options);
  }

  Future close() async{
    if (this.transport != null && this.clientOrDeviceHandle != null){
      return this.transport?.close(this.clientOrDeviceHandle).then((res){
        if (res.runtimeType != JRpcAndIosError){
          this.rootCapture = null;
          this.transport = null;
          this.clientOrDeviceHandle = null;
        } else {
          return res.toJson();
        }
      });
    } else {
      return JRpcAndIosError(0, -19, 'Transport is not initialized').toJson();
    }
  }

  openDevice(String guid, Capture capture) async {
    
    if(capture == null) {
      return JRpcAndIosError(0, SktErrors().ESKT_INVALIDPARAMETER, 'The provided parameter is invalid');
    } else {
      this.rootCapture = capture;
      this.transport = capture.transport;
      this.transportHandle = capture.transportHandle;

      var x = await this.transport?.openDevice(this.rootCapture!.clientOrDeviceHandle, guid).then((value) {
        this.clientOrDeviceHandle = value;
        return SktErrors().ESKT_NOERROR;
      });

      return x;

    }

  }

  Future getProperty(CaptureProperty property) async {
    if (this.transport != null) {
      return this.transport?.getProperty(this.clientOrDeviceHandle, property);
    } else {
      return JRpcAndIosError(0, -19, 'Transport is not initialized').toJson();
    }
  }

  Future setProperty(CaptureProperty property) async  {
    if (this.transport != null) {
      return this.transport?.setProperty(this.clientOrDeviceHandle, property);
    } else {
      return JRpcAndIosError(0, -19, 'Transport is not initialized').toJson();
    }

  }
  
}