// ignore_for_file: unnecessary_this
import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:http/http.dart';
import 'package:web_socket_channel/web_socket_channel.dart';
// import 'package:pigeon/pigeon.dart';
import './gen/errors.dart';
import './index.dart';

class HttpTransport extends Transport {
  String? host;
  bool? isWeb;
  Logger? logger;
  String hostWebsocket = '';
  WebSocketChannel? websocket;
  Function? onEventNotification;
  int rpcId = 1;

  HttpTransport(this.logger, [this.isWeb]);

  @override   
  Future getProperty(int? clientOrDeviceHandle, CaptureProperty property) async {

    var newParams = Params(clientOrDeviceHandle, property: property.toJson()).toJson();
    var req = JRpcRequest(
      this._getJsonRpcId(), 
      'getproperty', 
      newParams
    );
      
      return await this.send(this.transportHandle, req).then((response){
        if(response.toJson()['error'] != null){
          return JRpcAndIosError(0, response.toJson()['error']['code'], response.toJson()['error']['message']).toJson();
        } else {
          var res = json.decode(response.toJson()['result'])['result'];
          if(clientOrDeviceHandle != res['handle']) {
            return JRpcAndIosError(0, -11, 'The response handle does not match with the handle of the request').toJson();
          }
          return res['property'];
        }
      });
  }

  @override
  Future setProperty(int? clientOrDeviceHandle, CaptureProperty property) async {
    var newParams = Params(clientOrDeviceHandle, property: property.toJson()).toJson();
    
    var req = JRpcRequest(
      this._getJsonRpcId(), 
      'setproperty', 
      newParams
    );
    
    return await this.send(this.transportHandle, req).then((response){
      if(response.toJson()['error'] != null){
        return JRpcAndIosError(0, response.toJson()['error']['code'], response.toJson()['error']['message']).toJson();
      } else {
        var res = json.decode(response.toJson()['result'])['result'];
        if(clientOrDeviceHandle != res['handle']) {
          return JRpcAndIosError(0, -11, 'The response handle does not match with the handle of the request').toJson();
        }
        return res['property'];
      }
    });
  }

  @override
  Future openDevice(int? clientHandle, String guid) async {
    var openRequest = JRpcRequest(this._getJsonRpcId(), 'opendevice', Params(clientHandle, guid: guid));
    var result = await this.send(this.transportHandle, openRequest);
    
    if (result.toJson()['error'] != null) {
      return JRpcAndIosError(0, result.toJson()['error']['code'], result.toJson()['error']['message']).toJson();
    } else if (result.toJson()['result'] != null && json.decode(result.toJson()['result'])['result'] != null) {
      return json.decode(result.toJson()['result'])['result']['handle'];
    }

    return JRpcAndIosError(0, SktErrors().ESKT_NOTINITIALIZED, 'Trying to use an object that is not yet initialized');

  }

  @override
  Future open(String host, AppInfo appInfo, Function notification) async {

    var jsonRpc = JRpcRequest(this._getJsonRpcId(), 'openclient', AppInfo(
      appInfo.appId,
      appInfo.developerId,
      appInfo.appKey
    ));

    var newHandle = this.generateHandle();

    if (newHandle == 0) {
      var error = JRpcAndIosError(0, SktErrors().ESKT_INVALIDHANDLE, 'Invalid Handle').toJson();
      return error;
    }

    this.transportHandle = newHandle;

    RegExp exp = RegExp(r'(http|ftp|https):');
    this.host = host + '/Capture/v1/api';
    this.hostWebsocket = host.replaceAll(exp, 'ws:');

    this.onEventNotification = notification;

    var response = await this.send(newHandle, jsonRpc);

    if (response.runtimeType == JRpcAndIosError){
      return response.toJson();
    } else {
      var res = response.toJson();
      var decodedProperty = json.decode(res['result'])['result'];
      if(res['result'] != null && decodedProperty['handle'] != null) {
        return decodedProperty['handle'];
      } else {
        return JRpcAndIosError(0, SktErrors().ESKT_NOERROR, "No error.").toJson();
      }
    }

  }

  @override
  Future close(int? handle) async {
    var index = this.handles.indexWhere((h) => h.handle == handle);
    if (index == -1) {
      var error = JRpcAndIosError(0, SktErrors().ESKT_INVALIDHANDLE, 'Invalid Handle');
      return throw(error);
    }
    this.handles.removeAt(index);
    var jsonRpc = JRpcRequest(this._getJsonRpcId(), 'close', Params(handle));
    return await this.send(this.transportHandle, jsonRpc).then((response){
      if (response.runtimeType == JRpcAndIosError){
        return response;
      } else {
        return SktErrors().ESKT_NOERROR;
      }
    });
  }

  Future send(int? handle, JRpcRequest request) async {
    var promise = await sendRequest(request);
    return promise;
  }

  // future 
  Future sendRequest(request) async {
    //https://127.0.0.1
    //18481
    //need real response string
    String uriString = this.host ?? '';
    final uri = Uri.parse(uriString);
    final headers = {'Content-Type': 'application/json'};

    // ignore: prefer_function_declarations_over_variables
    var helper = (JRpcRequest jsonRpc) async {
      try {
        var jsonRpcString = json.encode(jsonRpc.toJson());
        logger!.log(jsonRpc.method.toString() + ' =>', jsonRpcString);
        var res = await post(
          uri, 
          headers: headers,
          body: jsonRpcString
        );
        var err = json.decode(res.body)['error'];
        if (err != null){
          return JRpcAndIosError(1, err['code'], err['message'] ?? 'no message');
        } else {
          return JRpcResponse(1, res.body);
        }
      } catch (err) {
        return JRpcAndIosError(0, -2100, err.toString());
      }
    };

    var response = await helper(request);

    logger!.log(request.method.toString() + ' <=', json.encode(response));
    
    if(request.method == 'openclient') {
      // we want to start the web service here if we can
      this.openWebSocket((event){
        if (response.runtimeType == JRpcResponse){
          var res = response as JRpcResponse;    // send a waitForEvent
          if (res.toJson()['result'] != null){
            var handle = json.decode(res.toJson()['result'])['result']['handle'];
            var waitForEvent = JRpcRequest(1, 'waitforcaptureevent', Params(handle)).toJson();
            var waitForEventString = json.encode(waitForEvent);
            //use sink.add to send data to server
            this.websocket!.sink.add(waitForEventString);
          }
          return;
        }
      });
     }
    return response;
  }

  openWebSocket(void Function(dynamic event) callback) {
    this.websocket = WebSocketChannel.connect(Uri.parse(this.hostWebsocket)); //, 'jsonRpc');
    this.logger!.log('websocket =>', 'open attempt');
    callback(this.websocket);
    this.websocket!.stream.listen(
        (dynamic message) {
          this.logger!.log('receiving something through the websocket:', message);
          var jsonRes = json.decode(message);
          if (jsonRes['result'] != null && jsonRes['result']['event'] != null) {
            var handle = jsonRes['result']['handle'];
            var type = jsonRes['result']['event']['type'];
            var id = jsonRes['result']['event']['id'];
            var value = jsonRes['result']['event']['value'];
            var evt = CaptureEvent(id, type, value, handle);
            this.notification(evt);
          } else if (jsonRes['error'] != null){
            this.logger!.log('Error in listener', jsonRes['error']);
          } else {
            this.logger!.log('Uh oh =>', 'No result returned');
          }
        },
        onDone: () {
          this.logger!.log('websocket =>', 'websocket closed!');
        },
        onError: (error) {
          if (error.hashCode == 1006){
            var err = CaptureEvent(CaptureEventIds.error, CaptureEventTypes.ulong, SktErrors().ESKT_SERVICENOTCOMMUNICATING);
            this.notification(err);
          } else {
            var err = CaptureEvent(CaptureEventIds.error, CaptureEventTypes.ulong, SktErrors().ESKT_SERVICENOTCOMMUNICATING);
            this.notification(err);
          }
        },
    );
  }

  _getJsonRpcId(){
    if (this.rpcId < 999){
       return this.rpcId++;
    } else {
       return this.rpcId = 1;
    }
  }

  notification(CaptureEvent eventBroh, [int? handle]) {
    var result = eventBroh.toJson()['result'] ?? eventBroh.toJson();
    if(eventBroh != null && this.onEventNotification != null){
      var event = result['event'] ?? result;
      var func = this.onEventNotification as Function;
      var hand = handle ?? result['handle'];
      func(event, hand);
    }
  }

}
